import cryptography.hazmat.primitives.asymmetric.rsa
import cryptography.hazmat.primitives.asymmetric.ec
import cryptography.hazmat.primitives.asymmetric.padding
import cryptography.hazmat.primitives.hashes

from .errors import WebAuthnError
from . import data


def load_cose_key(cose_key: dict):
    kty = cose_key.pop(1, None)
    if not kty:
        raise WebAuthnError("Invalid data")

    if kty == 3:
        alg = cose_key.pop(3, None)
        if alg and alg not in data.RSA_ALGORITHMS:
            raise WebAuthnError("Invalid data")

        n = cose_key.pop(-1, None)
        e = cose_key.pop(-2, None)
        if bool(cose_key):
            raise WebAuthnError("Invalid data")
        if not n or not e:
            raise WebAuthnError("Invalid data")

        try:
            n = int.from_bytes(n, byteorder='big')
            e = int.from_bytes(e, byteorder='big')
        except ValueError:
            raise WebAuthnError("Invalid data")

        pub_nums = cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicNumbers(n=n, e=e)
    elif kty == 2:
        crv = cose_key.pop(-1, None)
        alg = cose_key.pop(3, None)
        if not crv:
            raise WebAuthnError("Invalid data")

        if alg and alg not in data.ES_ALGORITHMS:
            raise WebAuthnError("Invalid data")

        x = cose_key.pop(-2, None)
        y = cose_key.pop(-3, None)
        if bool(cose_key):
            raise WebAuthnError("Invalid data")
        if not x or not y:
            raise WebAuthnError("Invalid data")

        if crv == 1:
            crv = cryptography.hazmat.primitives.asymmetric.ec.SECP256R1()
        elif crv == 2:
            crv = cryptography.hazmat.primitives.asymmetric.ec.SECP384R1()
        elif crv == 3:
            crv = cryptography.hazmat.primitives.asymmetric.ec.SECP521R1()
        elif crv == 8:
            crv = cryptography.hazmat.primitives.asymmetric.ec.SECP256K1()
        else:
            raise WebAuthnError("Invalid data")

        try:
            x = int.from_bytes(x, byteorder='big')
            y = int.from_bytes(y, byteorder='big')
        except ValueError:
            raise WebAuthnError("Invalid data")

        pub_nums = cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicNumbers(
            x=x, y=y, curve=crv
        )
    else:
        raise WebAuthnError("Invalid data")

    try:
        pub_key = pub_nums.public_key()
    except ValueError:
        raise WebAuthnError("Invalid data")

    return pub_key, alg


def jwt_alg_id_to_cose(jwt_alg_id: str) -> int:
    if jwt_alg_id == "ES256":
        return data.ES256
    elif jwt_alg_id == "ES384":
        return data.ES384
    elif jwt_alg_id == "ES512":
        return data.ES512
    elif jwt_alg_id == "RS256":
        return data.RS256
    elif jwt_alg_id == "RS384":
        return data.RS384
    elif jwt_alg_id == "RS512":
        return data.RS512
    elif jwt_alg_id == "PS256":
        return data.PS256
    elif jwt_alg_id == "PS384":
        return data.PS384
    elif jwt_alg_id == "PS512":
        return data.PS512


def verify_signature(pubkey, msg_data: bytes, signature: bytes, cose_alg_id: int):
    if isinstance(pubkey, cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey):
        if cose_alg_id == data.ES256:
            alg = cryptography.hazmat.primitives.asymmetric.ec.ECDSA(
                cryptography.hazmat.primitives.hashes.SHA256()
            )
        elif cose_alg_id == data.ES384:
            alg = cryptography.hazmat.primitives.asymmetric.ec.ECDSA(
                cryptography.hazmat.primitives.hashes.SHA384()
            )
        elif cose_alg_id == data.ES512:
            alg = cryptography.hazmat.primitives.asymmetric.ec.ECDSA(
                cryptography.hazmat.primitives.hashes.SHA512()
            )
        else:
            raise WebAuthnError("Invalid data")

        pubkey.verify(signature, msg_data, alg)
    elif isinstance(pubkey, cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey):
        if cose_alg_id == data.RS256:
            padding, alg = cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(), \
                           cryptography.hazmat.primitives.hashes.SHA256()
        elif cose_alg_id == data.RS384:
            padding, alg = cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(), \
                           cryptography.hazmat.primitives.hashes.SHA384()
        elif cose_alg_id == data.RS512:
            padding, alg = cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(), \
                           cryptography.hazmat.primitives.hashes.SHA512()
        elif cose_alg_id == data.PS256:
            padding, alg = cryptography.hazmat.primitives.asymmetric.padding.PSS(
                cryptography.hazmat.primitives.asymmetric.padding.MGF1(
                    cryptography.hazmat.primitives.hashes.SHA256()
                ), cryptography.hazmat.primitives.asymmetric.padding.PSS.MAX_LENGTH
            ), cryptography.hazmat.primitives.hashes.SHA256()
        elif cose_alg_id == data.PS384:
            padding, alg = cryptography.hazmat.primitives.asymmetric.padding.PSS(
                cryptography.hazmat.primitives.asymmetric.padding.MGF1(
                    cryptography.hazmat.primitives.hashes.SHA384()
                ), cryptography.hazmat.primitives.asymmetric.padding.PSS.MAX_LENGTH
            ), cryptography.hazmat.primitives.hashes.SHA384()
        elif cose_alg_id == data.PS512:
            padding, alg = cryptography.hazmat.primitives.asymmetric.padding.PSS(
                cryptography.hazmat.primitives.asymmetric.padding.MGF1(
                    cryptography.hazmat.primitives.hashes.SHA512()
                ), cryptography.hazmat.primitives.asymmetric.padding.PSS.MAX_LENGTH
            ), cryptography.hazmat.primitives.hashes.SHA512()
        else:
            raise WebAuthnError("Invalid data")

        pubkey.verify(signature, msg_data, padding, alg)


def key_equal(left, right):
    if isinstance(left, cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey) and \
            isinstance(right, cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey):
        left = left.public_numbers()
        right = right.public_numbers()
        return left == right
    elif isinstance(left, cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey) and \
            isinstance(right, cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey):
        left = left.public_numbers()
        right = right.public_numbers()
        return left == right
    else:
        return False



import logging
import enum
import dataclasses
import typing
import cryptography.x509
import uuid
import datetime
import base64
import urllib.parse
import certvalidator
import requests
import json
import binascii
import cryptography.hazmat.primitives.hashes
import cryptography.hazmat.primitives.asymmetric.rsa
import cryptography.hazmat.primitives.asymmetric.padding
import cryptography.exceptions

from . import data, errors

logger = logging.getLogger(__name__)


class FIDOCertification(enum.Enum):
    NotCertified = "NOT_FIDO_CERTIFIED"
    Certified = "FIDO_CERTIFIED"
    CertifiedL1 = "FIDO_CERTIFIED_L1"
    CertifiedL1Plus = "FIDO_CERTIFIED_L1plus"
    CertifiedL2 = "FIDO_CERTIFIED_L2"
    CertifiedL2Plus = "FIDO_CERTIFIED_L2plus"
    CertifiedL3 = "FIDO_CERTIFIED_L3"
    CertifiedL3Plus = "FIDO_CERTIFIED_L3plus"


@dataclasses.dataclass
class FIDOMetadataEntry:
    description: str
    protocol: str
    icon: typing.Optional[str]
    root_cas: typing.List[cryptography.x509.Certificate]
    fido_certification_level: FIDOCertification
    is_revoked: bool
    is_compromised: bool


@dataclasses.dataclass
class FIDOMetadata:
    entries: typing.List[FIDOMetadataEntry]
    aaguid_map: typing.Dict[uuid.UUID, int]
    cki_map: typing.Dict[str, int]

    @classmethod
    def from_metadata(cls, metadata):
        next_update = datetime.datetime.strptime(metadata["nextUpdate"], "%Y-%m-%d").date()
        today = datetime.datetime.utcnow().date()

        if next_update < today:
            logger.warning("Metadata out of date, please update")

        entries = []
        aaguid_map = {}
        cki_map = {}

        for entry in metadata["entries"]:
            root_cas = []
            statement = entry["metadataStatement"]

            fido_certification_level = FIDOCertification.NotCertified
            is_revoked = False
            is_compromised = False

            status_reports = sorted(filter(
                lambda d: d[0] <= today,
                map(
                    lambda d: (datetime.datetime.strptime(d["effectiveDate"], "%Y-%m-%d").date(), d),
                    entry["statusReports"]
                )
            ), key=lambda d: d[0])

            for _, report in status_reports:
                if report["status"] == FIDOCertification.NotCertified.value:
                    fido_certification_level = FIDOCertification.NotCertified
                elif report["status"] == FIDOCertification.Certified.value:
                    fido_certification_level = FIDOCertification.Certified
                elif report["status"] == FIDOCertification.CertifiedL1.value:
                    fido_certification_level = FIDOCertification.CertifiedL1
                elif report["status"] == FIDOCertification.CertifiedL1Plus.value:
                    fido_certification_level = FIDOCertification.CertifiedL1Plus
                elif report["status"] == FIDOCertification.CertifiedL2.value:
                    fido_certification_level = FIDOCertification.CertifiedL2
                elif report["status"] == FIDOCertification.CertifiedL2Plus.value:
                    fido_certification_level = FIDOCertification.CertifiedL2Plus
                elif report["status"] == FIDOCertification.CertifiedL3.value:
                    fido_certification_level = FIDOCertification.CertifiedL3
                elif report["status"] == FIDOCertification.CertifiedL3Plus.value:
                    fido_certification_level = FIDOCertification.CertifiedL3Plus
                elif report["status"] == "REVOKED":
                    is_revoked = True
                elif report["status"] == "USER_VERIFICATION_BYPASS":
                    is_compromised = True
                elif report["status"] == "ATTESTATION_KEY_COMPROMISE":
                    is_compromised = True
                elif report["status"] == "USER_KEY_REMOTE_COMPROMISE":
                    is_compromised = True
                elif report["status"] == "USER_KEY_PHYSICAL_COMPROMISE":
                    is_compromised = True

            if statement["schema"] != 3:
                continue

            for cert in statement["attestationRootCertificates"]:
                cert_bytes = base64.b64decode(cert)
                try:
                    root_cas.append(
                        cryptography.x509.load_der_x509_certificate(cert_bytes)
                    )
                except ValueError:
                    pass

            md_entry = FIDOMetadataEntry(
                description=statement["description"],
                protocol=statement["protocolFamily"],
                icon=statement.get("icon"),
                root_cas=root_cas,
                fido_certification_level=fido_certification_level,
                is_revoked=is_revoked,
                is_compromised=is_compromised,
            )
            index = len(entries)
            entries.append(md_entry)

            for cki in entry.get("attestationCertificateKeyIdentifiers", []):
                cki_map[cki] = index

            if "aaguid" in entry:
                aaguid_map[uuid.UUID(entry["aaguid"])] = index

        return cls(
            entries=entries,
            aaguid_map=aaguid_map,
            cki_map=cki_map
        )


CERT_START_LINE = b"-----BEGIN CERTIFICATE-----\n"
CERT_END_LINE = b"\n-----BEGIN CERTIFICATE-----"


def _get_signing_cert(root, h):
    validation_context = certvalidator.ValidationContext(
        trust_roots=[root], allow_fetching=True, revocation_mode="require"
    )

    if "x5u" in h:
        x5u = h["x5u"]
        if not isinstance(x5u, str):
            return None

        mds_parts = urllib.parse.urlsplit(data.FIDO_MDS_URL, scheme='https')
        x5u_parts = urllib.parse.urlsplit(x5u, scheme='https')
        if mds_parts.scheme != x5u_parts.scheme or mds_parts.netloc != x5u_parts.netloc:
            return None

        chain_r = requests.get(x5u, stream=True)
        chain_certs = list(map(
            lambda c: base64.b64decode(c.rstrip(CERT_END_LINE)),
            chain_r.iter_lines(delimiter=CERT_START_LINE)
        ))[1:]
    elif "x5c" in h:
        x5c = h["x5c"]
        if not isinstance(x5c, list):
            return None

        chain_certs = list(map(lambda c: base64.b64decode(c), x5c))
    else:
        return root

    validator = certvalidator.CertificateValidator(
        end_entity_cert=chain_certs[0],
        intermediate_certs=chain_certs[1:],
        validation_context=validation_context,
    )

    try:
        validator.validate_usage(
            key_usage={"digital_signature"},
            extended_optional=True
        )
    except certvalidator.errors.ValidationError:
        return None

    return cryptography.x509.load_der_x509_certificate(chain_certs[0])


def verify(cert, h, msg, sig):
    alg = h.get("alg")

    pubkey = cert.public_key()

    if alg in ("RS256", "RS384", "RS512"):
        if not isinstance(pubkey, cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey):
            logger.warning("Invalid key for RSA algorithm")
            return False

        if alg == "RS256":
            hasher = cryptography.hazmat.primitives.hashes.SHA256()
        elif alg == "RS384":
            hasher = cryptography.hazmat.primitives.hashes.SHA384()
        else:
            hasher = cryptography.hazmat.primitives.hashes.SHA512()

        try:
            pubkey.verify(
                sig,
                msg,
                cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(),
                hasher
            )
        except cryptography.exceptions.InvalidSignature:
            return False

        return True
    else:
        logger.warning("Unsupported signature algorithm")
        return False


def get_metadata():
    try:
        mds_r = requests.get(data.FIDO_MDS_URL)
    except requests.exceptions.RequestException:
        raise errors.WebAuthnError("Failed to fetch metadata")

    if mds_r.status_code != 200:
        raise errors.WebAuthnError("Failed to fetch metadata")

    mds_r.encoding = "ascii"
    try:
        mds_data = iter(mds_r.text.split("."))
        header_str = next(mds_data)
        header_bytes = base64.urlsafe_b64decode(header_str + '==')
        header = json.loads(header_bytes)
    except (json.JSONDecodeError, StopIteration, binascii.Error):
        raise errors.WebAuthnError("Invalid metadata")

    if header.get("typ") != "JWT":
        print("Metadata not a JWT")
        return

    cert = _get_signing_cert(data.FIDO_MDS_ROOT, header)

    if not cert:
        raise errors.WebAuthnError("Failed to get signing certificate")

    try:
        msg_str = next(mds_data)
        msg_bytes = base64.urlsafe_b64decode(msg_str + '==')
        sig_bytes = base64.urlsafe_b64decode(next(mds_data) + '==')
    except (StopIteration, binascii.Error):
        raise errors.WebAuthnError("Invalid metadata")

    if not verify(cert, header, (header_str + "." + msg_str).encode(), sig_bytes):
        raise errors.WebAuthnError("Metadata signature verification failed")

    try:
        msg = json.loads(msg_bytes)
    except json.JSONDecodeError:
        raise errors.WebAuthnError("Invalid metadata")

    return msg



import enum
import dataclasses
import typing
import base64
import cryptography.x509

from . import metadata


class AttestationType(enum.Enum):
    Self = enum.auto()
    Basic = enum.auto()
    AttestationCA = enum.auto()
    AnonymizationCA = enum.auto()
    NoneAttestation = enum.auto()


class AttestationMode(enum.Enum):
    Packed = "packed"
    TPM = "tpm"
    AndroidKey = "android-key"
    AndroidSafetynet = "android-safetynet"
    FIDOU2F = "fido-u2f"
    Apple = "apple"
    NoneAttestation = "none"


@dataclasses.dataclass
class AttestationResult:
    type: AttestationType
    mode: AttestationMode
    root_ca: typing.Optional[cryptography.x509.Certificate]
    cert_chain: typing.List[cryptography.x509.Certificate]
    safety_net_cts: typing.Optional[bool]
    fido_metadata: typing.Optional[metadata.FIDOMetadataEntry]


class AuthenticatorAttachment(enum.Enum):
    Platform = "platform"
    CrossPlatform = "cross-platform"


class UserVerification(enum.Enum):
    Required = "required"
    Preferred = "preferred"
    Discouraged = "discouraged"


class Attestation(enum.Enum):
    NoneAttestation = "none"
    IndirectAttestation = "indirect"
    DirectAttestation = "direct"


@dataclasses.dataclass
class RelyingParty:
    id: str
    name: str
    icon: typing.Optional[str]

    def to_options(self):
        out = {
            "id": self.id,
            "name": self.name
        }
        if self.icon:
            out["icon"] = self.icon
        return out


@dataclasses.dataclass
class User:
    id: bytes
    display_name: str
    name: str
    icon: typing.Optional[str]

    def to_options(self):
        out = {
            "id": base64.b64encode(self.id).decode(),
            "name": self.name,
            "displayName": self.display_name,
        }
        if self.icon:
            out["icon"] = self.icon
        return out



import typing
import os
import base64
import dataclasses
import uuid
import json
import cbor2
import binascii
import hashlib
import cryptography.hazmat.primitives.asymmetric.rsa
import cryptography.hazmat.primitives.asymmetric.ec
import cryptography.exceptions

from . import data, types, errors, utils, metadata, attestation


def create_webauthn_credentials(
        rp: types.RelyingParty, user: types.User, existing_keys: typing.List[bytes] = None,
        attachment: typing.Optional[types.AuthenticatorAttachment] = None, require_resident: bool = False,
        user_verification: typing.Optional[types.UserVerification] = None,
        attestation_request: typing.Optional[types.Attestation] = None
) -> typing.Tuple[dict, str]:
    if existing_keys is None:
        existing_keys = []

    if len(user.id) > 64:
        raise errors.WebAuthnError("User ID length is larger than limit of 64 bytes")

    challenge = os.urandom(64)
    challenge_base64 = base64.b64encode(challenge).decode()

    selection = {}
    if attachment:
        selection["authenticatorAttachment"] = attachment.value
    if require_resident:
        selection["requireResidentKey"] = bool(require_resident)
    if user_verification:
        selection["userVerification"] = user_verification.value

    options = {
        "rp": rp.to_options(),
        "user": user.to_options(),
        "challenge": challenge_base64,
        "pubKeyCredParams": [{
            "type": "public-key",
            "alg": alg
        } for alg in data.SUPPORTED_ALGORITHMS],
        "excludeCredentials": [{
            "type": "public-key",
            "id": base64.b64encode(k).decode()
        } for k in existing_keys],
        "authenticatorSelection": selection
    }
    if attestation:
        options["attestation"] = attestation_request.value

    return options, challenge_base64


def get_webauthn_credentials(
        rp: types.RelyingParty, existing_keys: typing.List[bytes] = None,
        user_verification: typing.Optional[types.UserVerification] = None,
) -> typing.Tuple[dict, str]:
    if existing_keys is None:
        existing_keys = []

    challenge = os.urandom(64)
    challenge_base64 = base64.b64encode(challenge).decode()

    options = {
        "rpId": rp.id,
        "challenge": challenge_base64,
        "allowCredentials": [{
            "type": "public-key",
            "id": base64.b64encode(k).decode()
        } for k in existing_keys],
    }
    if user_verification:
        options["userVerification"] = user_verification.value

    return options, challenge_base64


@dataclasses.dataclass
class CreateResult:
    aaguid: uuid.UUID
    sign_count: int
    public_key_alg: int
    public_key: typing.Union[
        cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey,
        cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey,
    ]
    attestation: types.AttestationResult


def verify_create_webauthn_credentials(
        rp: types.RelyingParty, challenge_b64: str, client_data_b64: str, attestation_b64: str,
        fido_metadata: metadata.FIDOMetadata, user_verification_required: bool = False
) -> CreateResult:
    try:
        challenge = base64.b64decode(challenge_b64)
        client_data_bytes = base64.b64decode(client_data_b64)
        client_data = json.loads(client_data_bytes.decode())
        attestation_data = cbor2.loads(base64.b64decode(attestation_b64))
    except (binascii.Error, ValueError, json.JSONDecodeError, cbor2.CBORDecodeError):
        raise errors.WebAuthnError("Invalid data")

    try:
        client_challenge = base64.urlsafe_b64decode(client_data["challenge"] + "==")
    except (binascii.Error, KeyError):
        raise errors.WebAuthnError("Invalid data")

    if client_challenge != challenge:
        raise errors.WebAuthnError("Verification failed")
    if client_data.get("type") != "webauthn.create":
        raise errors.WebAuthnError("Verification failed")
    if not client_data.get("origin", "").endswith(rp.id):
        raise errors.WebAuthnError("Verification failed")

    client_data_hash = hashlib.sha256(client_data_bytes).digest()
    rp_hash = hashlib.sha256(rp.id.encode()).digest()

    if "authData" not in attestation_data or type(attestation_data["authData"]) != bytes:
        raise errors.WebAuthnError("Invalid data")

    authenticator_data = attestation.AuthenticatorData.from_bytes(attestation_data["authData"])

    if rp_hash != authenticator_data.rp_hash:
        raise errors.WebAuthnError("Verification failed")

    if not authenticator_data.user_presence:
        raise errors.WebAuthnError("User not present")
    if not authenticator_data.user_verification and user_verification_required:
        raise errors.WebAuthnError("User not verified")

    verified_attestation = attestation.verify_attestation(
        attestation_statement=attestation_data, authenticator_data=authenticator_data,
        client_data_hash=client_data_hash, fido_metadata=fido_metadata
    )

    return CreateResult(
        aaguid=authenticator_data.attested_data.aaguid,
        sign_count=authenticator_data.sign_count,
        attestation=verified_attestation,
        public_key_alg=authenticator_data.attested_data.public_key_alg,
        public_key=authenticator_data.attested_data.public_key
    )


@dataclasses.dataclass
class GetResult:
    sign_count: int


def verify_get_webauthn_credentials(
        rp: types.RelyingParty, challenge_b64: str, client_data_b64: str, authenticator_b64: str, signature_b64: str,
        sign_count: int, pubkey_alg: int, pubkey, user_verification_required: bool = False,
) -> GetResult:
    try:
        challenge = base64.b64decode(challenge_b64)
        client_data_bytes = base64.b64decode(client_data_b64)
        client_data = json.loads(client_data_bytes.decode())
        authenticator = base64.b64decode(authenticator_b64)
        signature = base64.b64decode(signature_b64)
    except (binascii.Error, ValueError, json.JSONDecodeError):
        raise errors.WebAuthnError("Invalid data")

    authenticator_data = attestation.AuthenticatorData.from_bytes(authenticator)

    try:
        client_challenge = base64.urlsafe_b64decode(client_data["challenge"] + "==")
    except (binascii.Error, KeyError):
        raise errors.WebAuthnError("Invalid data")

    if client_challenge != challenge:
        raise errors.WebAuthnError("Verification failed")
    if client_data.get("type") != "webauthn.get":
        raise errors.WebAuthnError("Verification failed")
    if not client_data.get("origin", "").endswith(rp.id):
        raise errors.WebAuthnError("Verification failed")

    client_data_hash = hashlib.sha256(client_data_bytes).digest()
    rp_hash = hashlib.sha256(rp.id.encode()).digest()

    if rp_hash != authenticator_data.rp_hash:
        raise errors.WebAuthnError("Verification failed")

    if not authenticator_data.user_verification and user_verification_required:
        raise errors.WebAuthnError("User not verified")

    signed_data = authenticator + client_data_hash

    try:
        utils.verify_signature(pubkey, signed_data, signature, pubkey_alg)
    except cryptography.exceptions.InvalidSignature:
        raise errors.WebAuthnError("Verification failed")

    if sign_count != 0 and authenticator_data.sign_count != 0:
        if authenticator_data.sign_count > sign_count:
            sign_count = authenticator_data.sign_count
        else:
            raise errors.WebAuthnError("Verification failed")

    return GetResult(
        sign_count=sign_count
    )



import dataclasses
import uuid
import typing
import io
import struct
import cbor2
import json
import base64
import binascii
import hashlib
import certvalidator
import datetime
import cryptography.hazmat.primitives.asymmetric.rsa
import cryptography.hazmat.primitives.asymmetric.padding
import cryptography.hazmat.primitives.asymmetric.ec
import cryptography.hazmat.primitives.serialization
import cryptography.x509
import cryptography.exceptions

from .errors import WebAuthnError
from . import utils, types, data, metadata


@dataclasses.dataclass
class AttestedCredentialData:
    aaguid: uuid.UUID
    credential_id: bytes
    public_key_alg: int
    public_key: typing.Union[
        cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey,
        cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey,
    ]


@dataclasses.dataclass
class AuthenticatorData:
    rp_hash: bytes
    user_presence: bool
    user_verification: bool
    sign_count: int
    attested_data: typing.Optional[AttestedCredentialData]
    extensions: typing.Optional[dict]
    data_bytes: bytes

    @classmethod
    def from_bytes(cls, data_bytes: bytes):
        data = io.BytesIO(data_bytes)

        try:
            rp_hash, flags, sign_count = struct.unpack("!32sBI", data.read(37))
        except struct.error:
            raise WebAuthnError("Invalid data")
        up = bool(flags & 0b00000001)
        uv = bool(flags & 0b00000100)
        at = bool(flags & 0b01000000)
        ed = bool(flags & 0b10000000)

        if at:
            try:
                aaguid, cred_id_len = struct.unpack("!16sH", data.read(18))
            except struct.error:
                raise WebAuthnError("Invalid data")
            aaguid = uuid.UUID(bytes=aaguid)
            cred_id = data.read(cred_id_len)
            try:
                pubkey = cbor2.CBORDecoder(data).decode()
            except (IndexError, cbor2.CBORDecodeError):
                raise WebAuthnError("Invalid data")
            pubkey, alg = utils.load_cose_key(pubkey)

            attested_data = AttestedCredentialData(
                aaguid=aaguid,
                credential_id=cred_id,
                public_key_alg=alg,
                public_key=pubkey
            )
        else:
            attested_data = None

        if ed:
            try:
                extensions = cbor2.CBORDecoder(data).decode()
            except (IndexError, cbor2.CBORDecodeError):
                raise WebAuthnError("Invalid data")
        else:
            extensions = None

        return cls(
            rp_hash=rp_hash,
            user_presence=up,
            user_verification=uv,
            sign_count=sign_count,
            attested_data=attested_data,
            extensions=extensions,
            data_bytes=data_bytes
        )


def verify_attestation(
        attestation_statement: dict, authenticator_data: AuthenticatorData, client_data_hash: bytes,
        fido_metadata: metadata.FIDOMetadata
):
    fmt = attestation_statement.get("fmt", None)
    if not fmt:
        raise WebAuthnError("Invalid data")
    attestation_statement = attestation_statement.get("attStmt", None)
    if attestation_statement is None:
        raise WebAuthnError("Invalid data")

    if not authenticator_data.attested_data:
        raise WebAuthnError("Invalid data")

    signed_data = authenticator_data.data_bytes + client_data_hash

    if fmt == types.AttestationMode.Packed.value:
        alg = attestation_statement.get("alg", None)
        if alg not in data.SUPPORTED_ALGORITHMS:
            raise WebAuthnError("Invalid data")

        sig = attestation_statement.get("sig", None)
        if not sig:
            raise WebAuthnError("Invalid data")

        x5c = attestation_statement.get("x5c", None)

        if x5c:
            if len(x5c) < 1:
                raise WebAuthnError("Invalid data")

            try:
                certs = [cryptography.x509.load_der_x509_certificate(c) for c in x5c]
            except ValueError:
                raise WebAuthnError("Invalid data")
            except cryptography.exceptions.UnsupportedAlgorithm:
                raise WebAuthnError("Unsupported authenticator")

            out_certs = list(certs)

            attestation_cert: cryptography.x509.Certificate = certs.pop(0)

            try:
                utils.verify_signature(attestation_cert.public_key(), signed_data, sig, alg)
            except cryptography.exceptions.InvalidSignature:
                raise WebAuthnError("Verification failed")

            try:
                if attestation_cert.version != cryptography.x509.Version.v3:
                    raise WebAuthnError("Invalid data")
            except cryptography.x509.InvalidVersion:
                raise WebAuthnError("Invalid data")

            subject: cryptography.x509.Name = attestation_cert.subject
            subject_cc = subject.get_attributes_for_oid(cryptography.x509.NameOID.COUNTRY_NAME)
            subject_o = subject.get_attributes_for_oid(cryptography.x509.NameOID.ORGANIZATION_NAME)
            subject_ou = subject.get_attributes_for_oid(cryptography.x509.NameOID.ORGANIZATIONAL_UNIT_NAME)
            subject_cn = subject.get_attributes_for_oid(cryptography.x509.NameOID.COMMON_NAME)
            if len(subject_cc) != 1 or len(subject_cc[0].value) != 2:
                raise WebAuthnError("Verification failed")
            if len(subject_o) != 1:
                raise WebAuthnError("Verification failed")
            if len(subject_ou) != 1 or subject_ou[0].value != "Authenticator Attestation":
                raise WebAuthnError("Verification failed")
            if len(subject_cn) != 1:
                raise WebAuthnError("Verification failed")

            try:
                extensions = attestation_cert.extensions
            except (cryptography.x509.DuplicateExtension, cryptography.x509.UnsupportedGeneralNameType):
                raise WebAuthnError("Invalid data")

            try:
                basic_constraints = extensions.get_extension_for_oid(
                    cryptography.x509.ExtensionOID.BASIC_CONSTRAINTS
                )
                if basic_constraints.value.ca:
                    raise WebAuthnError("Verification failed")
            except cryptography.x509.ExtensionNotFound:
                raise WebAuthnError("Verification failed")

            try:
                aaguid_ext = extensions.get_extension_for_oid(
                    cryptography.x509.ObjectIdentifier("1.3.6.1.4.1.45724.1.1.4")
                )

                if aaguid_ext.critical:
                    raise WebAuthnError("Verification failed")

                aaguid = aaguid_ext.value.value
                if not aaguid.startswith(b"\x04\x10"):
                    raise WebAuthnError("Invalid data")
                aaguid = aaguid[2:]
                if len(aaguid) != 16:
                    raise WebAuthnError("Invalid data")
                aaguid = uuid.UUID(bytes=aaguid)

                if aaguid != authenticator_data.attested_data.aaguid:
                    raise WebAuthnError("Verification failed")
            except cryptography.x509.ExtensionNotFound:
                pass

            metadata_index = fido_metadata.aaguid_map.get(authenticator_data.attested_data.aaguid)
            if metadata_index is not None:
                metadata = fido_metadata.entries[metadata_index]

                if metadata.protocol != "fido2":
                    raise WebAuthnError("Verification failed")

                if metadata.is_revoked:
                    raise WebAuthnError("Verification failed")

                validator_context = certvalidator.ValidationContext(
                    trust_roots=[c.public_bytes(
                        cryptography.hazmat.primitives.serialization.Encoding.DER
                    ) for c in metadata.root_cas], allow_fetching=True
                )
                cert_validator = certvalidator.CertificateValidator(
                    end_entity_cert=x5c[0],
                    intermediate_certs=x5c[1:] if len(x5c) > 1 else None,
                    validation_context=validator_context
                )

                try:
                    path = cert_validator.validate_usage(key_usage=set())
                except certvalidator.errors.ValidationError as e:
                    raise WebAuthnError("Verification failed")

                path = [cryptography.x509.load_der_x509_certificate(c.dump()) for c in list(path)]

                return types.AttestationResult(
                    type=types.AttestationType.AttestationCA,
                    mode=types.AttestationMode.Packed,
                    root_ca=path[0],
                    cert_chain=path[1:],
                    safety_net_cts=None,
                    fido_metadata=metadata,
                )
            else:
                return types.AttestationResult(
                    type=types.AttestationType.Basic,
                    mode=types.AttestationMode.Packed,
                    root_ca=None,
                    cert_chain=out_certs,
                    safety_net_cts=None,
                    fido_metadata=None,
                )

        else:
            if alg != authenticator_data.attested_data.public_key_alg:
                raise WebAuthnError("Verification failed")

            try:
                utils.verify_signature(authenticator_data.attested_data.public_key, signed_data, sig, alg)
            except cryptography.exceptions.InvalidSignature:
                raise WebAuthnError("Verification failed")

            return types.AttestationResult(
                type=types.AttestationType.Self,
                mode=types.AttestationMode.Packed,
                root_ca=None,
                cert_chain=[],
                safety_net_cts=None,
                fido_metadata=None,
            )

    elif fmt == types.AttestationMode.AndroidSafetynet.value:
        try:
            response: bytes = attestation_statement.get("response", b"\xFF")
        except UnicodeDecodeError:
            raise WebAuthnError("Invalid data")

        if response.count(b".") != 2:
            raise WebAuthnError("Invalid data")

        jwt_sig_data, signature_str = response.rsplit(b".", 1)
        header_str, body_str = jwt_sig_data.split(b".", 1)
        try:
            header = json.loads(base64.urlsafe_b64decode(header_str + b"==").decode())
            body = json.loads(base64.urlsafe_b64decode(body_str + b"==").decode())
            signature = base64.urlsafe_b64decode(signature_str + b"==")
            certificates_bytes = [base64.b64decode(c) for c in header.get("x5c", [])]
            certificates = [cryptography.x509.load_der_x509_certificate(c) for c in certificates_bytes]
        except (ValueError, binascii.Error, json.JSONDecodeError):
            raise WebAuthnError("Invalid data")
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise WebAuthnError("Unsupported authenticator")

        if len(certificates) < 1:
            raise WebAuthnError("Invalid data")

        cert_validator = certvalidator.CertificateValidator(
            end_entity_cert=certificates_bytes[0],
            intermediate_certs=certificates_bytes[1:] if len(certificates_bytes) > 1 else None
        )
        try:
            path = list(cert_validator.validate_tls("attest.android.com"))
        except certvalidator.errors.ValidationError:
            raise WebAuthnError("Verification failed")

        root_ca = cryptography.x509.load_der_x509_certificate(path.pop(0).dump())
        cert_chain = [cryptography.x509.load_der_x509_certificate(c.dump()) for c in path]

        alg_id = utils.jwt_alg_id_to_cose(header.get("alg"))

        if alg_id is None:
            raise WebAuthnError("Unsupported authenticator")

        try:
            utils.verify_signature(certificates[0].public_key(), jwt_sig_data, signature, alg_id)
        except cryptography.exceptions.InvalidSignature:
            raise WebAuthnError("Verification failed")

        signed_data_b64 = base64.b64encode(hashlib.sha256(signed_data).digest()).decode()
        if body.get("nonce") != signed_data_b64:
            raise WebAuthnError("Verification failed")

        now = datetime.datetime.utcnow()
        timestamp = datetime.datetime.fromtimestamp(body.get("timestampMs", 0) / 1000.0)
        if now - timestamp > datetime.timedelta(minutes=5):
            raise WebAuthnError("Verification failed")

        if not body.get("basicIntegrity", False):
            raise WebAuthnError("Unsupported authenticator")

        return types.AttestationResult(
            type=types.AttestationType.Basic,
            mode=types.AttestationMode.AndroidSafetynet,
            root_ca=root_ca,
            cert_chain=cert_chain,
            safety_net_cts=body.get("ctsProfileMatch", False),
            fido_metadata=None
        )

    elif fmt == types.AttestationMode.FIDOU2F.value:
        x5c = attestation_statement.get("x5c", [])
        if len(x5c) != 1:
            raise WebAuthnError("Verification failed")

        sig = attestation_statement.get("sig", None)
        if not sig:
            raise WebAuthnError("Verification failed")

        try:
            cert = cryptography.x509.load_der_x509_certificate(x5c[0])
        except ValueError:
            raise WebAuthnError("Invalid data")
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise WebAuthnError("Unsupported authenticator")

        public_key = cert.public_key()
        if not isinstance(public_key, cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey):
            raise WebAuthnError("Verification failed")
        if public_key.curve.name != "secp256r1":
            raise WebAuthnError("Verification failed")

        cred_key_bytes = authenticator_data.attested_data.public_key.public_bytes(
            cryptography.hazmat.primitives.serialization.Encoding.X962,
            cryptography.hazmat.primitives.serialization.PublicFormat.UncompressedPoint
        )

        verification_data = bytearray([0x00])
        verification_data.extend(authenticator_data.rp_hash)
        verification_data.extend(client_data_hash)
        verification_data.extend(authenticator_data.attested_data.credential_id)
        verification_data.extend(cred_key_bytes)

        try:
            public_key.verify(sig, verification_data, cryptography.hazmat.primitives.asymmetric.ec.ECDSA(
                cryptography.hazmat.primitives.hashes.SHA256()
            ))
        except cryptography.exceptions.InvalidSignature:
            raise WebAuthnError("Verification failed")

        fingerprint = binascii.hexlify(
            cryptography.x509.SubjectKeyIdentifier.from_public_key(cert.public_key()).digest
        ).decode()
        metadata_index = fido_metadata.cki_map.get(fingerprint)

        if metadata_index is not None:
            metadata = fido_metadata.entries[metadata_index]

            if metadata.protocol != "u2f":
                raise WebAuthnError("Verification failed")

            if metadata.is_revoked:
                raise WebAuthnError("Verification failed")

            validator_context = certvalidator.ValidationContext(
                trust_roots=[c.public_bytes(
                    cryptography.hazmat.primitives.serialization.Encoding.DER
                ) for c in metadata.root_cas], allow_fetching=True
            )
            cert_validator = certvalidator.CertificateValidator(
                end_entity_cert=x5c[0],
                validation_context=validator_context,
            )

            try:
                path = cert_validator.validate_usage(key_usage=set())
            except certvalidator.errors.ValidationError:
                raise WebAuthnError("Verification failed")

            return types.AttestationResult(
                type=types.AttestationType.AttestationCA,
                mode=types.AttestationMode.FIDOU2F,
                root_ca=cryptography.x509.load_der_x509_certificate(path[0].dump()),
                cert_chain=[cert],
                safety_net_cts=None,
                fido_metadata=metadata,
            )

        return types.AttestationResult(
            type=types.AttestationType.Basic,
            mode=types.AttestationMode.FIDOU2F,
            root_ca=None,
            cert_chain=[cert],
            safety_net_cts=None,
            fido_metadata=None
        )

    elif fmt == types.AttestationMode.Apple.value:
        x5c = attestation_statement.get("x5c", [])
        if len(x5c) < 1:
            raise WebAuthnError("Invalid data")

        validator_context = certvalidator.ValidationContext(
            trust_roots=[data.APPLE_WEBAUTHN_ROOT]
        )
        cert_validator = certvalidator.CertificateValidator(
            end_entity_cert=x5c[0],
            intermediate_certs=x5c[1:] if len(x5c) > 1 else None,
            validation_context=validator_context
        )

        try:
            certs = [cryptography.x509.load_der_x509_certificate(c) for c in x5c]
        except ValueError:
            raise WebAuthnError("Invalid data")
        except cryptography.exceptions.UnsupportedAlgorithm:
            raise WebAuthnError("Unsupported authenticator")

        attestation_cert: cryptography.x509.Certificate = certs.pop(0)
        nonce = hashlib.sha256(signed_data).digest()

        try:
            if attestation_cert.version != cryptography.x509.Version.v3:
                raise WebAuthnError("Invalid data")
        except cryptography.x509.InvalidVersion:
            raise WebAuthnError("Invalid data")

        try:
            extensions = attestation_cert.extensions
        except (cryptography.x509.DuplicateExtension, cryptography.x509.UnsupportedGeneralNameType):
            raise WebAuthnError("Invalid data")

        try:
            nonce_ext = extensions.get_extension_for_oid(
                cryptography.x509.ObjectIdentifier("1.2.840.113635.100.8.2")
            )
        except cryptography.x509.ExtensionNotFound:
            raise WebAuthnError("Invalid data")

        cert_nonce = nonce_ext.value.value
        asn1_prefix = b"\x30\x24\xa1\x22\x04\x20"
        if not cert_nonce.startswith(asn1_prefix):
            raise WebAuthnError("Invalid data")

        if cert_nonce[len(asn1_prefix):] != nonce:
            raise WebAuthnError("Verification failed")

        if not utils.key_equal(authenticator_data.attested_data.public_key, attestation_cert.public_key()):
            raise WebAuthnError("Verification failed")

        try:
            path = cert_validator.validate_usage(
                key_usage={"digital_signature"}
            )
        except certvalidator.errors.ValidationError:
            raise WebAuthnError("Verification failed")

        path = [cryptography.x509.load_der_x509_certificate(c.dump()) for c in list(path)]

        return types.AttestationResult(
            type=types.AttestationType.AnonymizationCA,
            mode=types.AttestationMode.Apple,
            root_ca=path[0],
            cert_chain=path[1:],
            safety_net_cts=None,
            fido_metadata=None
        )

    elif fmt == types.AttestationMode.NoneAttestation.value:
        return types.AttestationResult(
            type=types.AttestationType.NoneAttestation,
            mode=types.AttestationMode.NoneAttestation,
            root_ca=None,
            cert_chain=[],
            safety_net_cts=None,
            fido_metadata=None
        )

    else:
        raise WebAuthnError("Unsupported authenticator")
